from contextlib import nullcontext
import pytest
from attrmagic import utils
from decimal import Decimal


@pytest.mark.parametrize(
    "value, expected",
    [
        (42, nullcontext(Decimal("42"))),
        (42.3, nullcontext(Decimal("42.3"))),
        ("42", nullcontext(Decimal("42"))),
        ("42.3", nullcontext(Decimal("42.3"))),
        (None, pytest.raises(ValueError)),
        ("hello", pytest.raises(ValueError)),
    ],
)
def test_coerce_to_dec(value, expected):
    with expected as e:
        assert utils.coerce_to_decimal(value) == e


@pytest.mark.parametrize(
    "value, expected",
    [
        (42, Decimal("42")),
        (42.3, Decimal("42.3")),
        ("42", Decimal("42")),
        ("42.3", Decimal("42.3")),
        ("hello", "hello"),
    ],
)
def test_decimal_or_string(value, expected):
    assert utils.decimal_or_string(value) == expected



"""Provides utilities for handling missing values using enums.

See: https://peps.python.org/pep-0484/#support-for-singleton-types-in-unions

Classes:
    Missing (enum.Enum): An enumeration to represent a missing value token.

Constants:
    MISSING (Missing): A constant representing the missing value token from the Missing enum.

__all__:
    List of public objects of this module, as interpreted by import *.

"""

from enum import Enum


class Missing(Enum):  # noqa: D101
    token = 0


MISSING = Missing.token



"""Operators for use in attrmagic.

These operators are used to compare values in attrmagic.

Classes:
    Operators: An enumeration of operators.

Functions:
    equals: Compare two values for equality.
    not_equal: Compare two values for inequality.
    greater_than: Compare two values for greater than.
    greater_than_or_equal: Compare two values for greater than or equal to.
    less_than: Compare two values for less than.
    less_than_or_equal_to: Compare two values for less than or equal to.
    in_: Check if value is in rhs.
    contains: Check if value contains rhs.
    icontains: Check if value contains rhs, case-insensitive.
    startswith: Check if value starts with rhs.
    istartswith: Check if value starts with rhs, case-insensitive.
    endswith: Check if value ends with rhs.
    iendswith: Check if value ends with rhs, case-insensitive.
    iequal: Check if value is equal to rhs, case-insensitive.
    range: Check if value is within a range.
"""

from collections.abc import Callable, Iterable
from datetime import datetime
from decimal import Decimal
from enum import Enum, member
from typing import TYPE_CHECKING, TypeVar, override

from pydantic import validate_call

from attrmagic.utils import validate_call_lex

_T = TypeVar("_T", Decimal, float, str)


@validate_call_lex
def equals(value: _T, rhs: _T) -> bool:
    """Compare two values for equality."""
    return value == rhs


@validate_call_lex
def not_equal(value: _T, rhs: _T) -> bool:
    """Compare two values for inequality."""
    return value != rhs


@validate_call_lex
def greater_than(value: _T, rhs: _T):
    """Compare two values for greater than.

    Example:
    ```
    >>> greater_than(5, 3)
    True
    >>> greater_than(3, 5)
    False
    ```

    Args:
        value: The value to compare [left-hand side]
        rhs: The value to compare against [right-hand side]
    """
    return value > rhs


@validate_call_lex
def greater_than_or_equal(value: _T, rhs: _T) -> bool:
    """Compare two values for greater than or equal to.

    Example:
    ```
    >>> greater_than_or_equal(5, 3)
    True
    >>> greater_than_or_equal(3, 5)
    False
    >>> greater_than_or_equal(3, 3)
    True
    ```

    Args:
        value: The value to compare [left-hand side]
        rhs: The value to compare against [right-hand side]
    """
    return value >= rhs


@validate_call_lex
def less_than(value: _T, rhs: _T):
    """Compare two values for less than.

    Example:
    ```
    >>> less_than(3, 5)
    True
    >>> less_than(5, 3)
    False
    ```

    Args:
        value: The value to compare [left-hand side]
        rhs: The value to compare against [right-hand side]
    """
    return value < rhs


@validate_call_lex
def less_than_or_equal_to(value: _T, rhs: _T):
    """Compare two values for less than or equal to.

    Example:
    ```
    >>> less_than_or_equal_to(3, 5)
    True
    >>> less_than_or_equal_to(5, 3)
    False
    >>> less_than_or_equal_to(3, 3)
    True
    ```

    Args:
        value: The value to compare [left-hand side]
        rhs: The value to compare against [right-hand side]
    """
    return value <= rhs


# @validate_call
def in_[T: (Decimal, float, str)](value: T, rhs: "Iterable[T]") -> bool:
    """Check if value is in rhs.

    Example:
    ```
    >>> in_(3, [1, 2, 3, 4, 5])
    True
    >>> in_(6, [1, 2, 3, 4, 5])
    False
    ```

    Args:
        value: The value to check
        rhs: The iterable to check against
    """
    return value in rhs


# TODO: should this accept more types?
@validate_call
def contains(value: str, rhs: str) -> bool:
    """Check if value contains rhs.

    Example:
    ```
    >>> contains("hello", "he")
    True
    >>> contains("hello", "hi")
    False
    ```

    Args:
        value: The value to check
        rhs: The value to check against
    """
    return rhs in value


@validate_call
def icontains(value: str, rhs: str) -> bool:
    """Check if value contains rhs, case-insensitive.

    Example:
    ```
    >>> icontains("hello", "he")
    True
    >>> icontains("hello", "Hi")
    False
    >>> icontains("HeLlO", "he")
    True
    ```

    Args:
        value: The value to check
        rhs: The value to check against
    """
    return rhs.lower() in value.lower()


@validate_call
def startswith(value: str, rhs: str) -> bool:
    """Check if value starts with rhs.

    Example:
    ```
    >>> startswith("hello", "he")
    True
    >>> startswith("the", "he")
    False
    ```

    Args:
        value: The value to check
        rhs: The value to check against
    """
    return value.startswith(rhs)


@validate_call
def istartswith(value: str, rhs: str) -> bool:
    """Check if value starts with rhs, case-insensitive.

    Example:
    ```
    >>> istartswith("hello", "he")
    True
    >>> istartswith("the", "He")
    False
    >>> istartswith("HELLO", "he")
    True
    ```

    Args:
        value: The value to check
        rhs: The value to check against
    """
    return value.lower().startswith(rhs.lower())


@validate_call
def endswith(value: str, rhs: str) -> bool:
    """Check if value ends with rhs.

    Example:
    ```
    >>> endswith("hello", "lo")
    True
    >>> endswith("the", "he")
    True
    >>> endswith("the", "th")
    False
    >>> endswith("the", "Th")
    False
    ```

    Args:
        value: The value to check
        rhs: The value to check against
    """
    return value.endswith(rhs)


@validate_call
def iendswith(value: str, rhs: str):
    """Check if value ends with rhs, case-insensitive.

    Example:
    ```
    >>> iendswith("hello", "LO")
    True
    >>> iendswith("the", "HE")
    True
    >>> iendswith("the", "TH")
    False
    >>> iendswith("the", "Th")
    True
    ```

    Args:
        value: The value to check
        rhs: The value to check against
    """
    return value.lower().endswith(rhs.lower())


@validate_call
def iequal(value: str, rhs: str):
    """Check if value is equal to rhs, case-insensitive.

    Example:
    ```
    >>> iequal("hello", "HELLO")
    True
    >>> iequal("the", "THE")
    True
    >>> iequal("the", "TH")
    False
    >>> iequal("the", "Th")
    False
    ```

    Args:
        value: The value to check
        rhs: The value to check against
    """
    return value.lower() == rhs.lower()


R = TypeVar("R", datetime, Decimal, float, int, str)


@validate_call
def range(value: R, rhs: tuple[R, R]) -> bool:
    """Check if value is within a range.

    Example:
    ```
    >>> range(datetime(2021, 1, 1), (datetime(2020, 1, 1), datetime(2022, 1, 1)))
    True
    >>> range(datetime(2023, 1, 1), (datetime(2020, 1, 1), datetime(2022, 1, 1)))
    False
    ```

    Args:
        value: The value to check
        rhs: The range to check against
    """
    return rhs[0] <= value <= rhs[1]


class Operators(Enum):
    """An enumeration of operators."""

    EXACT = member(equals)
    IEXACT = member(iequal)
    CONTAINS = member(contains)
    ICONTAINS = member(icontains)
    IN = member(in_)
    GT = member(greater_than)
    GTE = member(greater_than_or_equal)
    LT = member(less_than)
    LTE = member(less_than_or_equal_to)
    NE = member(not_equal)
    STARTSWITH = member(startswith)
    ISTARTSWITH = member(istartswith)
    ENDSWITH = member(endswith)
    IENDSWITH = member(iendswith)
    RANGE = member(range)
    EQUAL = member(equals)
    IEQUAL = member(iequal)

    if TYPE_CHECKING:

        @property
        @override
        def value(self) -> Callable[[_T, _T], bool]:
            return self._value_  # pyright: ignore[reportReturnType]

    # TODO: Typing is probably not totally correct here
    def evaluate(self, value: _T, rhs: _T) -> bool:
        """Evaluate the operator.

        Args:
            value: The value to compare
            rhs: The value to compare against

        Returns:
            bool: The result of the comparison
        """
        return self.value(value, rhs)



"""Provides utilities for attribute path manipulation and error handling.

Classes:
    AttrPath: A class for handling attribute paths.

Functions:
    getattr_path: Get an attribute path.
    setattr_path: Set an attribute path.

Exceptions:
    UnknownValueError: An exception for unknown values.
"""

from collections import deque
from collections.abc import Generator
from typing import Any, SupportsIndex, cast, override

from pydantic import BaseModel, model_validator

from attrmagic.operators import Operators

from .sentinels import MISSING, Missing
from .utils import path_as_parts


def getattr_path[T](
    obj: object,
    path: "str | AttrPath",
    *,
    separator: str = "__",
    default: T | Missing = MISSING,
) -> object | T:
    """Get an attribute path, as defined by a string separated by '__'.

    Example:
    ```
    >>> foo = Foo(a=Bar(b=Baz(c=42)))
    >>> getattr_path(foo, 'a__b__c')
    42
    ```


    Args:
        obj: The object to get the attribute from.
        path: The path to the attribute.
        separator: The separator to use.
        default: The default value to return if the attribute is not found.

    Returns:
        The attribute at the given path.

    Raises:
        AttributeError: If the attribute does not exist, including any intermediate attributes.
    """
    if path == "":
        return obj
    current = obj
    attr_path = AttrPath.str_to_path(path, separator=separator)
    for name in attr_path:
        if default is MISSING:
            try:
                current = getattr(current, name)  # pyright: ignore[reportAny]
            except AttributeError as e:
                msg = f"'{type(obj).__name__}' object has no attribute path '{path}', since {e}"
                raise AttributeError(msg) from e

        else:
            current = getattr(current, name, MISSING)
            if current is MISSING:
                return default
        if current is None:
            return None
    return current


def setattr_path(
    obj: object, path: str, value: object, *, separator: str = "__"
) -> None:
    """Set an attribute path, as defined by a string separated by '__'.

    Example:
    ```
    >>> foo = Foo(a=Bar(b=Baz(c=42)))
    >>> setattr_path(foo, 'a__b__c', 43)
    >>> foo.a.b.c
    43
    ```

    Args:
        obj: The top-level object to set the attribute on.
        path: The path to the attribute.
        value: The value to set the attribute to.
        separator: The separator to use.
    """
    current = obj
    parts = path_as_parts(path, separator=separator)
    for name in list(parts)[:-1]:
        current = getattr(current, name)  # pyright: ignore[reportAny]
    setattr(current, parts[-1], value)


class AttrPath(BaseModel):
    """Represents a path-like structure using a string with a specified separator.

    Attributes:
        value (str): The string representation of the path.

    Methods:
        str_to_path(cls, value: "str | AttrPath") -> "AttrPath": Converts a string or AttrPath instance to an AttrPath instance.
        parts(): Returns the deque of parts.
        pop(index: int = -1): Removes and returns the part at the specified index.
        popleft(): Removes and returns the leftmost part.
        get(index: int): Returns the part at the specified index.
        render(): Joins the parts into a string using the separator and returns it.
        depth: Returns the number of parts in the path.
    """

    separator: str = "__"
    value: str
    _parts: deque[str] | None = None

    @model_validator(mode="before")
    @classmethod
    def value_as_string(cls, data: Any) -> Any:  # pyright: ignore[reportAny, reportExplicitAny]
        """Converts a sequence of values into a path string."""
        if isinstance(data, dict):
            if isinstance(data["value"], str):
                return data  # pyright: ignore[reportUnknownVariableType]

            separator = cast(
                "str",
                (
                    data["separator"]
                    if "separator" in data
                    else cls.get_field_default("separator")
                ),
            )
            data["value"] = separator.join(data["value"])  # pyright: ignore[reportUnknownArgumentType]
        return data  # pyright: ignore[reportUnknownVariableType]

    @classmethod
    def get_field_default(cls, field_name: str) -> Any:  # pyright: ignore[reportAny, reportExplicitAny]
        """Returns the default value for the specified field."""
        fields = cls.model_fields
        field = next(fields[name] for name in fields if name == field_name)
        return field.default  # pyright: ignore[reportAny]

    @classmethod
    def str_to_path(
        cls, value: "str | AttrPath", *, separator: str = "__"
    ) -> "AttrPath":
        """Converts a string or AttrPath instance to an AttrPath instance.

        Args:
            value: The string or AttrPath instance to convert. If an AttrPath instance is passed, it is returned as is.
            separator: The separator to use.

        Returns:
            An AttrPath instance.
        """
        if not isinstance(value, AttrPath):
            return cls(value=value, separator=separator)
        return value

    @override
    def __iter__(self) -> Generator[str, Any, None]:  # pyright: ignore[reportExplicitAny, reportIncompatibleMethodOverride]
        """Returns an iterator over the parts of the path."""
        yield from iter(self.parts)

    def __getitem__(self, index: SupportsIndex):
        """Returns the part at the specified index."""
        return self.parts[index]

    @property
    def parts(self) -> deque[str]:
        """Returns the parts attribute.

        Returns:
            list: The parts attribute.
        """
        if self._parts is None:
            self._parts = path_as_parts(self.value, separator=self.separator)
        return self._parts

    def pop(self, index: int = -1) -> str:
        """Removes and returns the part at the specified index."""
        parts = self.parts
        parts.rotate(-index)
        value = parts.popleft()
        parts.rotate(index)
        return value

    def popleft(self) -> str:
        """Removes and returns the leftmost part."""
        return self.parts.popleft()

    def render(self):
        """Joins the parts into a string using the separator and returns it."""
        return self.separator.join(self.parts)

    @override
    def __str__(self) -> str:
        return self.render()

    @property
    def depth(self):
        """Returns the number of parts in the path."""
        return len(self.parts)


class QueryPath(BaseModel):
    """Represents a query path."""

    attr_path: AttrPath
    operator: Operators = Operators.EXACT
    _separator: str = "__"

    @classmethod
    def from_string(cls, value: str, *, separator: str = "__") -> "QueryPath":
        """Creates a QueryPath instance from a string.

        Args:
            value: The string to convert.
            separator: The separator to use.

        Returns:
            A QueryPath instance.
        """
        parts = list(path_as_parts(value, separator=separator))

        operator_candidate = parts[-1].upper()
        if operator_candidate not in Operators.__members__:
            path = separator.join(parts)
            return cls(
                attr_path=AttrPath(value=path, separator=separator),
                _separator=separator,
            )
        path = separator.join(parts[:-1])
        return cls(
            attr_path=AttrPath(value=path, separator=separator),
            operator=Operators[operator_candidate],
            _separator=separator,
        )



from collections import deque
from contextlib import nullcontext
from typing import Any

import pytest
from polyfactory.factories.pydantic_factory import ModelFactory

from attrmagic import core
from attrmagic.sentinels import MISSING


def test_getattr_path():
    class Foo:
        def __init__(self, a):
            self.a = a

    foo = Foo(a=42)
    assert core.getattr_path(foo, "a") == 42
    with pytest.raises(AttributeError):
        core.getattr_path(foo, "b")


class Bar:
    def __init__(self, b):
        self.b = b


@pytest.fixture(scope="function")
def bar(request):
    return Bar(b=request.param)


def test_blank_path():
    bar_local = Bar(b=42)
    assert core.getattr_path(bar_local, "") == bar_local


@pytest.mark.parametrize(
    ("bar", "path", "default", "expected"),
    [
        (42, "a", MISSING, pytest.raises(AttributeError)),
        (42, "a", None, nullcontext(None)),
        (42, "a", 43, nullcontext(43)),
        (42, "b", None, nullcontext(42)),
        (42, "b", MISSING, nullcontext(42)),
        (None, "b", MISSING, nullcontext(None)),
    ],
    indirect=["bar"],
)
def test_param_getattr_path(bar: Bar, path: str, default: Any, expected):
    with expected as e:
        assert core.getattr_path(bar, path, default=default) == e


def test_getattr_nested_path():
    path = "b__b"
    bar = Bar(b=42)
    foo = Bar(b=bar)
    assert core.getattr_path(foo, path) == 42
    assert core.getattr_path(foo, "b") == bar
    with pytest.raises(AttributeError):
        assert core.getattr_path(foo, "a")


@pytest.fixture
def bar_default():
    return Bar(b=42)


def test_setattr_path_simple(bar_default: Bar):
    core.setattr_path(bar_default, "b", 43)
    assert bar_default.b == 43
    core.setattr_path(bar_default, "b", 44)
    assert bar_default.b == 44


def test_setattr_path_nested():
    bar_child = Bar(b=42)
    bar_parent = Bar(b=bar_child)

    core.setattr_path(bar_parent, "b__b", 43)
    assert bar_child.b == 43


class AttrPathFactory(ModelFactory[core.AttrPath]):
    separator = "__"
    value = "a__b__c"


def test_attr_path_creation():
    attr_path = AttrPathFactory.build()
    assert isinstance(attr_path, core.AttrPath)
    assert isinstance(attr_path.value, str)
    assert attr_path.value == "a__b__c"
    assert attr_path.depth == 3
    assert attr_path.parts == deque(["a", "b", "c"])
    first = attr_path.popleft()
    assert first == "a"
    assert attr_path.render() == "b__c"

    assert str(attr_path) == "b__c"

    assert attr_path[1] == "c"
    assert attr_path.pop() == "c"


def test_pop():
    attr_path = AttrPathFactory.build()
    assert attr_path.pop(1) == "b"
    assert attr_path.render() == "a__c"


def test_attr_str_to_path():
    path_str = "a__b__c"
    attr_path = core.AttrPath.str_to_path(path_str)

    assert isinstance(attr_path, core.AttrPath)
    assert isinstance(attr_path.value, str)
    assert attr_path.value == path_str
    assert attr_path.depth == 3
    assert attr_path.parts == deque(["a", "b", "c"])


def test_attr_from_parts():
    parts = ["a", "b", "c"]
    attr_path = core.AttrPath(value=parts)

    assert isinstance(attr_path, core.AttrPath)
    assert isinstance(attr_path.value, str)
    assert attr_path.value == "a__b__c"
    assert attr_path.depth == 3

    assert str(attr_path) == "a__b__c"


def test_attr_path_to_path():
    path_str = AttrPathFactory.build()

    attr_path = core.AttrPath.str_to_path(path_str)

    assert attr_path is path_str


def test_query_path():
    query_path = core.QueryPath.from_string("a__b__c")
    assert query_path.attr_path.render() == "a__b__c"
    assert query_path.operator == core.Operators.EXACT
    assert query_path._separator == "__"
    assert query_path.attr_path.separator == "__"


def test_query_path_iexact():
    query_path = core.QueryPath.from_string("a__b__c__iexact")
    assert query_path.attr_path.render() == "a__b__c"
    assert query_path.operator == core.Operators.IEXACT
    assert query_path._separator == "__"
    assert query_path.attr_path.separator == "__"



"""Provides utility functions for manipulating path strings and coercing values to lists.

Functions:
    - path_as_parts: Convert a path string to a list of parts.
    - get_path_part: Get a part of a path string.
    - path_popleft: Remove the first part of a path string.
    - path_popright: Remove the last part of a path string.
    - coerce_to_list: Ensure that a value is a list.
"""

import functools
from collections import deque
from collections.abc import Callable
from decimal import Decimal, InvalidOperation
from typing import ParamSpec, TypeVar


def path_as_parts(path: str, *, separator: str = "__") -> deque[str]:
    """Convert a path string to a list of parts.

    Args:
      path: The path string.
      separator: The separator to use.

    Returns:
        A list of parts.

    Examples:
        >>> path_as_parts("a__b__c")
        ["a", "b", "c"]

    """
    return deque(path.split(separator))


LEX_TYPES_PRIORITY = [Decimal, int, float, str]

LexType = Decimal | int | float | str


def coerce_to_decimal(value: LexType) -> Decimal:
    """Coerce a value to a Decimal.

    Args:
        value: The value to coerce.

    Returns:
        Decimal: The coerced value.

    Raises:
        ValueError: If the value cannot be coerced.
    """
    if isinstance(value, Decimal):
        return value
    try:
        return Decimal(str(value))
    except InvalidOperation:
        raise ValueError(f"Cannot coerce value {value} to Decimal.") from None


def decimal_or_string(value: LexType) -> Decimal | str:
    """Coerce a value to a Decimal or a string.

    Args:
        value: The value to coerce.

    Returns:
        Decimal | str: The coerced value.

    Raises:
        ValueError: If the value cannot be coerced.
    """
    try:
        return coerce_to_decimal(value)
    except ValueError:
        return str(value)


ARGS = TypeVar("ARGS", tuple[Decimal | float | str], dict[str, Decimal | float | str])
ValueT = TypeVar("ValueT", Decimal, float, str)
RHS_T = TypeVar("RHS_T", Decimal, float, str)
# T = TypeVar("T")

_ParamsT = ParamSpec("_ParamsT")
_ReturnT = TypeVar("_ReturnT")


def validate_call_lex[**ParamsT, ReturnT](
    func: Callable[ParamsT, ReturnT],
) -> Callable[ParamsT, ReturnT]:
    """Validate the call of a function with lexical arguments.

    Args:
        func: The function to validate

    Returns:
        Callable: The validated function
    """

    @functools.wraps(func)
    def wrapper(*args: ParamsT.args, **_kwargs: ParamsT.kwargs) -> ReturnT:
        args = (decimal_or_string(arg) for arg in args)  # pyright: ignore[reportAssignmentType, reportArgumentType]
        return func(*args)  # pyright: ignore[reportCallIssue]

    return wrapper



"""Defines models for attribute-based searching and filtering using Pydantic BaseModel.

Classes:
    ClassBase: A base Pydantic class that adds the ability to get attributes by path.
    SearchBase: A generic root model for searching and filtering lists of ClassBase objects.
"""

from collections.abc import Iterable, Iterator, Mapping, Sequence
from decimal import Decimal
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Literal,
    Self,
    SupportsIndex,
    cast,
    overload,
    override,
)

from pydantic import BaseModel, RootModel
from pydantic_core.core_schema import ListSchema, ModelSchema

from attrmagic.core import AttrPath, QueryPath, getattr_path
from attrmagic.operators import Operators
from attrmagic.sentinels import MISSING, Missing

if TYPE_CHECKING:
    from _collections_abc import dict_items, dict_keys, dict_values


class ClassBase(BaseModel):
    """Base pydantic class that adds the ability to get attributes by path."""

    def getattr_path[T](
        self,
        attr_path: str | AttrPath,
        *,
        separator: str = "__",
        default: T | Missing = MISSING,
    ) -> object | T:
        """Get an attribute path, as defined by a string separated by '__'.

        Example:
        ```
        >>> foo = Foo(a=Bar(b=Baz(c=42)))
        >>> foo.getattr_path('a__b__c')
        42
        ```

        Args:
            attr_path: The path to the attribute.
            separator: The separator to use.
            default: The default value to return if the attribute is not found.

        Returns:
            The attribute at the given path.

        Raises:
            AttributeError: If the attribute does not exist, including any intermediate attributes.
        """
        return getattr_path(
            obj=self, path=attr_path, separator=separator, default=default
        )


class SimpleBaseGenericList[SimpleBase](list[SimpleBase]): ...  # noqa: D101


class SimpleBaseGenericDict[SimpleBase](dict[str, SimpleBase]): ...  # noqa: D101


class Filter[SimpleBase](BaseModel):
    """A filter that can be applied to a list of objects."""

    path: QueryPath
    value: object

    @cached_property
    def attr_path(self) -> AttrPath:
        """Get the attribute path."""
        return self.path.attr_path

    @cached_property
    def operator(self) -> Operators:
        """Get the operator."""
        return self.path.operator

    @classmethod
    def from_kwarg(cls, **kwargs: object) -> list[Self]:
        """Create a filter from a kwarg."""

        def qp(path: str) -> QueryPath:
            return QueryPath.from_string(path)

        return [cls(path=qp(path), value=value) for path, value in kwargs.items()]

    def evaluate(self, item: SimpleBase) -> bool:
        """Evaluate the filter against an item."""
        value = cast("Decimal | float | str", getattr_path(item, self.attr_path))
        return self.operator.evaluate(value, self.value)  # pyright: ignore[reportArgumentType]


def _get_or_raise[T](obj: Mapping[str, T], attr: str) -> T:
    result = obj[attr]
    if result is None:
        raise AttributeError(f"Attribute '{attr}' not found in {obj}")
    return result


class SimpleListRoot[SimpleBase](RootModel[list[SimpleBase]]):  # noqa: D101
    @classmethod
    def empty(cls) -> Self:
        """Create an empty instance of the class."""
        return cls(root=[])

    @override
    def __iter__(self) -> Iterator[SimpleBase]:  # pyright: ignore[reportIncompatibleMethodOverride]
        return iter(self.root)

    @overload
    def __getitem__(self, item: SupportsIndex, /) -> SimpleBase: ...
    @overload
    def __getitem__(self, item: slice, /) -> list[SimpleBase]: ...
    def __getitem__(self, item: SupportsIndex | slice) -> SimpleBase | list[SimpleBase]:  # noqa: D105
        return self.root[item]

    @overload
    def __setitem__(self, key: SupportsIndex, value: SimpleBase) -> None: ...
    @overload
    def __setitem__(self, key: slice, value: Sequence[SimpleBase]) -> None: ...
    def __setitem__(  # noqa: D105
        self, key: SupportsIndex | slice, value: SimpleBase | Sequence[SimpleBase]
    ) -> None:
        if isinstance(key, SupportsIndex) and not isinstance(value, Sequence):
            self.root[key] = value
        elif isinstance(key, slice) and isinstance(value, Sequence):
            self.root[key] = value
        else:
            raise TypeError(
                f"Invalid types for __setitem__: key={type(key)}, value={type(value)}"  # pyright: ignore[reportUnknownArgumentType]
            )

    def _get_filters(self, **kwargs: object) -> list[Filter[SimpleBase]]:
        return Filter[SimpleBase].from_kwarg(**kwargs)

    def _filter_list(self, filters: Iterable[Filter[SimpleBase]]) -> Self:
        assert isinstance(self.root, list), "_filter_list requires that root is a list"
        for filter in filters:
            self.root: list[SimpleBase] = [
                item for item in self.root if filter.evaluate(item)
            ]
        return self

    @property
    def base_type(self) -> type[SimpleBase]:
        """Get the base type of the items in the root list."""
        core_schema = self.__class__.__pydantic_core_schema__
        schema = cast("ListSchema", _get_or_raise(core_schema, "schema"))
        items_schema = cast("ModelSchema", _get_or_raise(schema, "items_schema"))
        return cast("type[SimpleBase]", _get_or_raise(items_schema, "cls"))

    def filter(self, **kwargs: object) -> Self:
        """Find items that match the kwargs.

        Example:
        ```
        >>> search = SimpleRoot[list[Foo]](root=[Foo(a=1), Foo(a=2), Foo(a=3)])
        >>> search.filter(a__gt=1)
        SearchRoot([Foo(a=2), Foo(a=3)])
        >>> search = SimpleRoot[list[int]](root=[1, 2, 3])
        >>> search.filter(gt=1)
        SimpleRoot([2, 3])
        ```

        Args:
            kwargs: The attributes to filter by.
        """
        filters = self._get_filters(**kwargs)
        assert isinstance(self.root, list), "root must be a list"
        return self._filter_list(filters)

    @overload
    def get(
        self, *, default: SimpleBase | Missing = MISSING, **kwargs: object
    ) -> SimpleBase: ...
    @overload
    def get(self, *, default: Literal[None], **kwargs: object) -> SimpleBase | None: ...
    def get(
        self,
        *,
        default: SimpleBase | Missing | None = MISSING,
        **kwargs: object,
    ) -> SimpleBase | None:
        """Return the item that matches the kwargs or the default value.

        Raises:
            ValueError: If 0 or more than 1 items are returned.
        """
        items_list = self.model_copy().filter(**kwargs)

        if (list_len := len(items_list)) != 1:
            if default is MISSING:
                match list_len:
                    case 0:
                        msg = "get() returned no items"
                    case _:
                        msg = "get() returned more than one item"
                raise ValueError(msg)
            return default

        return items_list[0]

    def pop(self, index: SupportsIndex = -1, /) -> SimpleBase:
        """Remove and return item at index (default last)."""
        return self.root.pop(index)

    def append(self, item: SimpleBase):
        """Append an item to the end of class."""
        self.root.append(item)

    def __add__(  # noqa: D105
        self, other: "SimpleListRoot[SimpleBase] | Iterable[SimpleBase] | object"
    ) -> Self:
        match other:
            case SimpleListRoot():
                self.root += other.root  # pyright: ignore[reportUnknownMemberType]
            case Iterable():
                self.root += list(other)  # pyright: ignore[reportUnknownArgumentType]
            case _:
                return NotImplemented

        return self

    def __len__(self) -> int:  # noqa: D105
        return len(self.root)

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.root})"


class SimpleDict[KeyT, ValueT](RootModel[dict[KeyT, ValueT]]):
    """An implementation of Pydantic's BaseModel for dictionaries.

    Adds (most) methods from the built-in dict class.
    """

    @classmethod
    def empty(cls) -> Self:
        """Create an empty instance of the class."""
        return cls(root={})

    def __len__(self) -> int:
        """Return the number of items in the dictionary."""
        return len(self.root)

    def keys(self) -> "dict_keys[KeyT, ValueT]":
        """Return a new view of the dictionary's keys."""
        return self.root.keys()

    def values(self) -> "dict_values[KeyT, ValueT]":
        """Return a new view of the dictionary's values."""
        return self.root.values()

    def items(self) -> "dict_items[KeyT, ValueT]":
        """Return a new view of the dictionary's items (key, value)."""
        return self.root.items()

    @overload  # type: ignore[override]
    def get(self, key: KeyT, /) -> ValueT | None: ...
    @overload
    def get(self, key: KeyT, default: ValueT, /) -> ValueT: ...
    @overload
    def get[T](self, key: KeyT, default: T, /) -> ValueT | T: ...
    def get[T](
        self, key: KeyT, default: T | ValueT | Missing = MISSING, /
    ) -> ValueT | T | None:
        """Return the value for key if key is in the dictionary, else default.

        Args:
            key: The key to get.
            default: The default value to return if the key is not found.
        """
        normal_default = default if default is not MISSING else None

        return self.root.get(key, normal_default)

    @overload
    def pop(self, key: KeyT, /) -> ValueT: ...
    @overload
    def pop(self, key: KeyT, default: ValueT, /) -> ValueT: ...
    @overload
    def pop[T](self, key: KeyT, default: T, /) -> ValueT | T: ...
    def pop[T](
        self, key: KeyT, default: T | ValueT | Missing = MISSING, /
    ) -> ValueT | T:
        """Remove specified key and return the corresponding value.

        Args:
            key: The key to remove.
            default: The default value to return if the key is not found.
        """
        if default is MISSING:
            return self.root.pop(key)
        return self.root.pop(key, default)

    def __getitem__(self, key: KeyT) -> ValueT:
        """Return the value for key."""
        return self.root[key]

    def __setitem__(self, key: KeyT, value: ValueT) -> None:
        """Set the value for key."""
        self.root[key] = value

    def __delitem__(self, key: KeyT) -> None:
        """Delete self[key]."""
        del self.root[key]

    @override
    def __eq__(self, value: object) -> bool:
        """Return self==value."""
        return self.root == value

    def __reversed__(self) -> Iterator[KeyT]:
        """Return a reverse iterator over the keys of the dictionary."""
        return reversed(self.root)

    def __contains__(self, key: KeyT) -> bool:
        """Return key in self."""
        return key in self.root


class SimpleDictStr[ValueT](RootModel[dict[str, ValueT]]):  # noqa: D101
    pass


class SearchBase[SearchRoot: ClassBase](SimpleListRoot[SearchRoot]):
    """A generic root model for searching and filtering lists of ClassBase objects.

    Example:
    ```
    >>> class Foo(ClassBase):
    ...     a: int
    ...
    >>> search = SearchBase([Foo(a=1), Foo(a=2), Foo(a=3)])
    >>> search.filter(a__gt=1)
    SearchBase([Foo(a=2), Foo(a=3)])
    ```
    """

    def _compare(
        self, item: SearchRoot, lhs: str | AttrPath, rhs: object, operator: Operators
    ) -> bool:
        value = item.getattr_path(attr_path=lhs)
        return operator.evaluate(value, rhs)  # pyright: ignore[reportArgumentType]

    def _split_kwarg[T](self, **kwargs: T) -> tuple[str, T]:
        """Return tuple of lhs and rhs."""
        assert len(kwargs) <= 1, "only one kwarg is allowed beyond default"

        return next(iter(kwargs.items()))

    def _get_compare_tuple[T](self, **kwargs: T) -> tuple[AttrPath, T, Operators]:
        """Return tuple of lhs, rhs, and operator."""
        lhs, rhs = self._split_kwarg(**kwargs)
        query_path = QueryPath.from_string(lhs)
        lhs, operator = query_path.attr_path, query_path.operator

        return lhs, rhs, operator

    def exclude(self, **kwargs: object) -> Self:
        """Remove items that match the kwargs."""
        lhs, rhs, operator = self._get_compare_tuple(**kwargs)

        self.root: list[SearchRoot] = [
            item for item in self.root if not self._compare(item, lhs, rhs, operator)
        ]

        return self



from contextlib import nullcontext

import pytest

from attrmagic import operators
from attrmagic.operators import R
from datetime import datetime as dt
from decimal import Decimal


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        (42, 42, nullcontext(True)),
        (42, 43, nullcontext(False)),
        (42, "42", nullcontext(True)),
        ("42", "42", nullcontext(True)),
        ("42", 42, nullcontext(True)),
        ("42", "43", nullcontext(False)),
        (42.3, 42.3, nullcontext(True)),
        (42.3, 42.4, nullcontext(False)),
        (42.3, "42.3", nullcontext(True)),
        ("42.3", "42.3", nullcontext(True)),
        ("42.3", 42.3, nullcontext(True)),
        ("42.3", "42.4", nullcontext(False)),
        (None, None, nullcontext(True)),
        ("hello", "hello", nullcontext(True)),
        ("hello", 45, nullcontext(False)),
        ("hElLo", "hello", nullcontext(False)),
    ],
)
def test_equals(value, rhs, expected):
    with expected as e:
        assert operators.equals(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        ("hElLo", "hello", nullcontext(True)),
        ("hello", "hello", nullcontext(True)),
        ("hElL", "hello", nullcontext(False)),
    ],
)
def test_iequal(value, rhs, expected):
    with expected as e:
        assert operators.iequal(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        (42, 42, nullcontext(False)),
        (42, 43, nullcontext(True)),
        (42, "42", nullcontext(False)),
        ("42", "42", nullcontext(False)),
        ("42", 42, nullcontext(False)),
        ("42", "43", nullcontext(True)),
        (42.3, 42.3, nullcontext(False)),
        (42.3, 42.4, nullcontext(True)),
        (42.3, "42.3", nullcontext(False)),
        ("42.3", "42.3", nullcontext(False)),
        ("42.3", 42.3, nullcontext(False)),
        ("42.3", "42.4", nullcontext(True)),
        (None, None, nullcontext(False)),
        ("hello", "hello", nullcontext(False)),
        ("hello", 45, nullcontext(True)),
    ],
)
def test_not_equal(value, rhs, expected):
    with expected as e:
        assert operators.not_equal(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        (42, 42, nullcontext(False)),
        (42, 43, nullcontext(False)),
        (43, 42, nullcontext(True)),
        (Decimal("43.1"), Decimal("42.9"), nullcontext(True)),
    ],
)
def test_greater_than(value, rhs, expected):
    with expected as e:
        assert operators.greater_than(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        (42, 42, nullcontext(True)),
        (42, 43, nullcontext(False)),
        (43, 42, nullcontext(True)),
    ],
)
def test_greater_than_or_equal(value, rhs, expected):
    with expected as e:
        assert operators.greater_than_or_equal(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        (42, 42, nullcontext(False)),
        (42, 43, nullcontext(True)),
        (43, 42, nullcontext(False)),
    ],
)
def test_less_than(value, rhs, expected):
    with expected as e:
        assert operators.less_than(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        (42, 42, nullcontext(True)),
        (42, 43, nullcontext(True)),
        (43, 42, nullcontext(False)),
    ],
)
def test_less_than_or_equal(value, rhs, expected):
    with expected as e:
        assert operators.less_than_or_equal_to(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        (42, [1, 42, 10_000], nullcontext(True)),
        (43, [1, 42, 10_000], nullcontext(False)),
    ],
)
def test_in(value, rhs, expected):
    with expected as e:
        assert operators.in_(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        ("hello, world", ", wo", nullcontext(True)),
        ("hello, world", ", WO", nullcontext(False)),
        ("goodbye", ", wo", nullcontext(False)),
    ],
)
def test_contains(value, rhs, expected):
    with expected as e:
        assert operators.contains(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        ("hello, world", ", wo", nullcontext(True)),
        ("hello, world", ", WO", nullcontext(True)),
        ("goodbye", ", wo", nullcontext(False)),
    ],
)
def test_icontains(value, rhs, expected):
    with expected as e:
        assert operators.icontains(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        ("hello, world", "hell", nullcontext(True)),
        ("hello, world", "Hell", nullcontext(False)),
        ("goodbye", "ood", nullcontext(False)),
    ],
)
def test_startswith(value, rhs, expected):
    with expected as e:
        assert operators.startswith(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        ("hello, world", "hell", nullcontext(True)),
        ("hello, world", "Hell", nullcontext(True)),
        ("goodbye", "ood", nullcontext(False)),
    ],
)
def test_istartswith(value, rhs, expected):
    with expected as e:
        assert operators.istartswith(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        ("hello, world", "world", nullcontext(True)),
        ("hello, world", "World", nullcontext(False)),
        ("goodbye", "ood", nullcontext(False)),
    ],
)
def test_endswith(value, rhs, expected):
    with expected as e:
        assert operators.endswith(value, rhs) == e


@pytest.mark.parametrize(
    "value, rhs, expected",
    [
        ("hello, world", "world", nullcontext(True)),
        ("hello, world", "World", nullcontext(True)),
        ("goodbye", "ood", nullcontext(False)),
    ],
)
def test_iendswith(value, rhs, expected):
    with expected as e:
        assert operators.iendswith(value, rhs) == e


@pytest.mark.parametrize(
    ("value", "rhs", "expected"),
    [
        (42, (1, 100), nullcontext(True)),
        (42, (1, 41), nullcontext(False)),
        (42, (43, 100), nullcontext(False)),
        (dt(2021, 1, 1), (dt(2021, 1, 1), dt(2021, 12, 31)), nullcontext(True)),
        (dt(2021, 1, 1), (dt(2021, 1, 2), dt(2021, 12, 31)), nullcontext(False)),
    ],
)
def test_range(value: R, rhs: tuple[R, R], expected):
    with expected as e:
        assert operators.range(value, rhs) == e



import enum

import pytest

from attrmagic.models import ClassBase, SearchBase, SimpleDict


class Bar(ClassBase):
    c: int


class Foo(ClassBase):
    a: Bar


class MyTestEnum(enum.Enum):
    A = 1
    B = 2
    C = 3


class SimpleDictTest(SimpleDict[str, MyTestEnum]): ...


# class BarSearch(SearchBase[Bar]):
#     pass

BarSearch = SearchBase[Bar]


@pytest.fixture
def bar_search():
    return BarSearch([Bar(c=1), Bar(c=2), Bar(c=3)])


def test_getattr_path():
    foo = Foo(a=Bar(c=42))
    assert foo.getattr_path("a__c") == 42
    assert foo.getattr_path("a__d", default="missing") == "missing"
    with pytest.raises(AttributeError):
        _ = foo.getattr_path("a__d")


def test_searchbase_filter(bar_search: BarSearch):
    filtered = bar_search.filter(c__gt=1)
    assert len(filtered) == 2
    assert filtered[0].c == 2
    assert filtered[1].c == 3


def test_searchbase_exclude(bar_search: BarSearch):
    excluded = bar_search.exclude(c__gt=1)
    assert len(excluded) == 1
    assert excluded[0].c == 1


def test_searchbase_append(bar_search: BarSearch):
    bar_search_copy = bar_search.model_copy()

    bar_search_copy.append(Bar(c=2))
    assert len(bar_search_copy) == 4
    assert bar_search_copy[3].c == 2


def test_searchbase_add(bar_search: BarSearch):
    new_search = BarSearch([Bar(c=4)])

    search1 = bar_search
    search2 = new_search
    combined = search1 + search2
    assert len(combined) == 4
    assert combined[0].c == 1
    assert combined[3].c == 4


def test_searchbase_add_list(bar_search: BarSearch):
    new_search = [Bar(c=4)]

    search1 = bar_search
    search2 = new_search
    combined = search1 + search2
    assert len(combined) == 4
    assert combined[0].c == 1
    assert combined[3].c == 4


def test_searchbase_add_number(bar_search: BarSearch):
    with pytest.raises(TypeError):
        bar_search + 1  # pyright: ignore[reportUnusedExpression]


def test_searchbase_iter(bar_search: BarSearch):
    assert list(bar_search) == [Bar(c=1), Bar(c=2), Bar(c=3)]


def test_searchbase_setitem(bar_search: BarSearch):
    bar_search[0] = Bar(c=4)
    assert bar_search[0].c == 4


def test_searchbase_get(bar_search: BarSearch):
    assert bar_search.get(c__exact=2) == bar_search[1]
    with pytest.raises(ValueError):
        assert bar_search.get(c__exact=4) is None
    assert bar_search.get(c__exact=4, default=bar_search[0]) == bar_search[0]
    assert bar_search.get(c=2) == bar_search[1]
    assert bar_search.get(c=2, c__exact=2) == bar_search[1]
    assert bar_search.get(c__exact=4, default=None) is None


def test_repr(bar_search: BarSearch):
    assert repr(bar_search) == "SearchBase[Bar]([Bar(c=1), Bar(c=2), Bar(c=3)])"


def test_simple_dict():
    my_simple_dict = SimpleDictTest(root={"a": MyTestEnum.A, "b": MyTestEnum.B})

    assert my_simple_dict["a"] == MyTestEnum.A
